from posixpath import join as pjoin

import pandas as pd
import pytest
from pandas.util.testing import assert_frame_equal

import ibis
import ibis.util as util
from ibis.tests.util import assert_equal

pytest.importorskip('hdfs')
pytest.importorskip('sqlalchemy')
pytest.importorskip('impala.dbapi')

impala = pytest.importorskip('impala')

from ibis.impala.compat import ImpylaError  # noqa: E402, isort:skip

pytestmark = pytest.mark.impala


@pytest.fixture
def df():
    df = pd.DataFrame(
        {
            'year': [2009] * 3 + [2010] * 3,
            'month': list(map(str, [1, 2, 3] * 2)),
            'value': list(range(1, 7)),
        },
        index=list(range(6)),
    )
    df = pd.concat([df] * 10, ignore_index=True)
    df['id'] = df.index.values
    return df


@pytest.fixture
def unpart_t(con, df, tmp_db):
    pd_name = '__ibis_test_partition_{}'.format(util.guid())
    con.create_table(pd_name, df, database=tmp_db)
    try:
        yield con.table(pd_name, database=tmp_db)
    finally:
        assert con.exists_table(pd_name, database=tmp_db), pd_name
        con.drop_table(pd_name, database=tmp_db)


def test_is_partitioned(con, temp_table):
    schema = ibis.schema(
        [('foo', 'string'), ('year', 'int32'), ('month', 'string')]
    )
    name = temp_table
    con.create_table(name, schema=schema, partition=['year', 'month'])
    assert con.table(name).is_partitioned


def test_create_table_with_partition_column(con, temp_table_db):
    schema = ibis.schema(
        [
            ('year', 'int32'),
            ('month', 'string'),
            ('day', 'int8'),
            ('value', 'double'),
        ]
    )

    tmp_db, name = temp_table_db
    con.create_table(
        name, schema=schema, database=tmp_db, partition=['year', 'month']
    )

    # the partition column get put at the end of the table
    ex_schema = ibis.schema(
        [
            ('day', 'int8'),
            ('value', 'double'),
            ('year', 'int32'),
            ('month', 'string'),
        ]
    )
    table_schema = con.get_schema(name, database=tmp_db)
    assert_equal(table_schema, ex_schema)

    partition_schema = con.database(tmp_db).table(name).partition_schema()

    expected = ibis.schema([('year', 'int32'), ('month', 'string')])
    assert_equal(partition_schema, expected)


def test_create_partitioned_separate_schema(con, temp_table):
    schema = ibis.schema([('day', 'int8'), ('value', 'double')])
    part_schema = ibis.schema([('year', 'int32'), ('month', 'string')])

    name = temp_table
    con.create_table(name, schema=schema, partition=part_schema)

    # the partition column get put at the end of the table
    ex_schema = ibis.schema(
        [
            ('day', 'int8'),
            ('value', 'double'),
            ('year', 'int32'),
            ('month', 'string'),
        ]
    )
    table_schema = con.get_schema(name)
    assert_equal(table_schema, ex_schema)

    partition_schema = con.table(name).partition_schema()
    assert_equal(partition_schema, part_schema)


def test_unpartitioned_table_get_schema(con):
    tname = 'functional_alltypes'
    with pytest.raises(ImpylaError):
        con.table(tname).partition_schema()


def test_insert_select_partitioned_table(con, df, temp_table, unpart_t):
    part_keys = ['year', 'month']

    con.create_table(temp_table, schema=unpart_t.schema(), partition=part_keys)
    part_t = con.table(temp_table)
    unique_keys = df[part_keys].drop_duplicates()

    for i, (year, month) in enumerate(unique_keys.itertuples(index=False)):
        select_stmt = unpart_t[
            (unpart_t.year == year) & (unpart_t.month == month)
        ]

        # test both styles of insert
        if i:
            part = {'year': year, 'month': month}
        else:
            part = [year, month]
        part_t.insert(select_stmt, partition=part)

    verify_partitioned_table(part_t, df, unique_keys)


def test_create_partitioned_table_from_expr(con, alltypes):
    t = alltypes
    expr = t[t.id <= 10][['id', 'double_col', 'month', 'year']]
    name = 'tmppart_{}'.format(util.guid())
    try:
        con.create_table(name, expr, partition=[t.year])
    except Exception:
        raise
    else:
        new = con.table(name)
        expected = expr.execute().sort_values('id').reset_index(drop=True)
        result = new.execute().sort_values('id').reset_index(drop=True)
        assert_frame_equal(result, expected)
    finally:
        con.drop_table(name, force=True)


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_insert_overwrite_partition():
    assert False


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_dynamic_partitioning():
    assert False


def test_add_drop_partition_no_location(con, temp_table):
    schema = ibis.schema(
        [('foo', 'string'), ('year', 'int32'), ('month', 'int16')]
    )
    name = temp_table
    con.create_table(name, schema=schema, partition=['year', 'month'])
    table = con.table(name)

    part = {'year': 2007, 'month': 4}

    table.add_partition(part)

    assert len(table.partitions()) == 2

    table.drop_partition(part)

    assert len(table.partitions()) == 1


def test_add_drop_partition_owned_by_impala(hdfs, con, temp_table):
    schema = ibis.schema(
        [('foo', 'string'), ('year', 'int32'), ('month', 'int16')]
    )
    name = temp_table
    con.create_table(name, schema=schema, partition=['year', 'month'])

    table = con.table(name)

    part = {'year': 2007, 'month': 4}

    subdir = util.guid()
    basename = util.guid()
    path = '/tmp/{}/{}'.format(subdir, basename)

    hdfs.mkdir('/tmp/{}'.format(subdir))
    hdfs.chown('/tmp/{}'.format(subdir), owner='impala', group='supergroup')

    table.add_partition(part, location=path)

    assert len(table.partitions()) == 2

    table.drop_partition(part)

    assert len(table.partitions()) == 1


def test_add_drop_partition_hive_bug(con, temp_table):
    schema = ibis.schema(
        [('foo', 'string'), ('year', 'int32'), ('month', 'int16')]
    )
    name = temp_table
    con.create_table(name, schema=schema, partition=['year', 'month'])

    table = con.table(name)

    part = {'year': 2007, 'month': 4}

    path = '/tmp/{}'.format(util.guid())

    table.add_partition(part, location=path)

    assert len(table.partitions()) == 2

    table.drop_partition(part)

    assert len(table.partitions()) == 1


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_set_partition_location():
    assert False


def test_load_data_partition(con, hdfs, tmp_dir, unpart_t, df, temp_table):
    part_keys = ['year', 'month']

    con.create_table(temp_table, schema=unpart_t.schema(), partition=part_keys)
    part_t = con.table(temp_table)

    # trim the runtime of this test
    df = df[df.month == '1'].reset_index(drop=True)

    unique_keys = df[part_keys].drop_duplicates()

    hdfs_dir = pjoin(tmp_dir, 'load-data-partition')

    df2 = df.drop(['year', 'month'], axis='columns')

    csv_props = {'serialization.format': ',', 'field.delim': ','}

    for i, (year, month) in enumerate(unique_keys.itertuples(index=False)):
        chunk = df2[(df.year == year) & (df.month == month)]
        chunk_path = pjoin(hdfs_dir, '{}.csv'.format(i))

        con.write_dataframe(chunk, chunk_path)

        # test both styles of insert
        if i:
            part = {'year': year, 'month': month}
        else:
            part = [year, month]

        part_t.add_partition(part)
        part_t.alter_partition(part, format='text', serde_properties=csv_props)
        part_t.load_data(chunk_path, partition=part)

    hdfs.rmdir(hdfs_dir)
    verify_partitioned_table(part_t, df, unique_keys)


def verify_partitioned_table(part_t, df, unique_keys):
    result = (
        part_t.execute()
        .sort_values(by='id')
        .reset_index(drop=True)[df.columns]
    )

    assert_frame_equal(result, df)

    parts = part_t.partitions()

    # allow for the total line
    assert len(parts) == len(unique_keys) + 1


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_drop_partition():
    assert False


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_repartition_automated():
    assert False



import datetime
import operator

import pytest

import ibis
import ibis.expr.api as api
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir


def test_temporal_literals():
    date = ibis.literal('2015-01-01', 'date')
    assert isinstance(date, ir.DateScalar)

    timestamp = ibis.literal('2017-01-01 12:00:33', 'timestamp')
    assert isinstance(timestamp, ir.TimestampScalar)


@pytest.mark.parametrize(
    ('interval', 'unit', 'expected'),
    [
        (api.interval(months=3), 'Q', api.interval(quarters=1)),
        (api.interval(months=12), 'Y', api.interval(years=1)),
        (api.interval(quarters=8), 'Y', api.interval(years=2)),
        (api.interval(days=14), 'W', api.interval(weeks=2)),
        (api.interval(minutes=240), 'h', api.interval(hours=4)),
        (api.interval(seconds=360), 'm', api.interval(minutes=6)),
        (api.interval(seconds=3 * 86400), 'D', api.interval(days=3)),
        (api.interval(milliseconds=5000), 's', api.interval(seconds=5)),
        (api.interval(microseconds=5000000), 's', api.interval(seconds=5)),
        (api.interval(nanoseconds=5000000000), 's', api.interval(seconds=5)),
    ],
)
def test_upconvert(interval, unit, expected):
    result = interval.to_unit(unit)

    assert isinstance(result, ir.IntervalScalar)
    assert result.type().unit == expected.type().unit


@pytest.mark.parametrize('target', ['Y', 'Q', 'M'])
@pytest.mark.parametrize(
    'delta',
    [
        api.interval(weeks=1),
        api.interval(days=1),
        api.interval(hours=1),
        api.interval(minutes=1),
        api.interval(seconds=1),
        api.interval(milliseconds=1),
        api.interval(microseconds=1),
        api.interval(nanoseconds=1),
    ],
)
def test_cannot_upconvert(delta, target):
    with pytest.raises(ValueError):
        delta.to_unit(target)


@pytest.mark.parametrize(
    'expr',
    [
        api.interval(days=2) * 2,
        api.interval(days=2) * (-2),
        3 * api.interval(days=2),
        (-3) * api.interval(days=2),
    ],
)
def test_multiply(expr):
    assert isinstance(expr, ir.IntervalScalar)
    assert expr.type().unit == 'D'


@pytest.mark.parametrize(
    'expr',
    [
        api.interval(days=1) + api.interval(days=1),
        api.interval(days=2) + api.interval(hours=4),
    ],
)
def test_add(expr):
    assert isinstance(expr, ir.IntervalScalar)
    assert expr.type().unit == 'D'


@pytest.mark.parametrize(
    'expr',
    [
        api.interval(days=3) - api.interval(days=1),
        api.interval(days=2) - api.interval(hours=4),
    ],
)
def test_subtract(expr):
    assert isinstance(expr, ir.IntervalScalar)
    assert expr.type().unit == 'D'


@pytest.mark.parametrize(
    ('case', 'expected'),
    [
        (api.interval(seconds=2).to_unit('s'), api.interval(seconds=2)),
        (
            api.interval(seconds=2).to_unit('ms'),
            api.interval(milliseconds=2 * 1000),
        ),
        (
            api.interval(seconds=2).to_unit('us'),
            api.interval(microseconds=2 * 1000000),
        ),
        (
            api.interval(seconds=2).to_unit('ns'),
            api.interval(nanoseconds=2 * 1000000000),
        ),
        (
            api.interval(milliseconds=2).to_unit('ms'),
            api.interval(milliseconds=2),
        ),
        (
            api.interval(milliseconds=2).to_unit('us'),
            api.interval(microseconds=2 * 1000),
        ),
        (
            api.interval(milliseconds=2).to_unit('ns'),
            api.interval(nanoseconds=2 * 1000000),
        ),
        (
            api.interval(microseconds=2).to_unit('us'),
            api.interval(microseconds=2),
        ),
        (
            api.interval(microseconds=2).to_unit('ns'),
            api.interval(nanoseconds=2 * 1000),
        ),
        (
            api.interval(nanoseconds=2).to_unit('ns'),
            api.interval(nanoseconds=2),
        ),
    ],
)
def test_downconvert_second_parts(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.type().unit == expected.type().unit


@pytest.mark.parametrize(
    ('case', 'expected'),
    [
        (api.interval(hours=2).to_unit('h'), api.interval(hours=2)),
        (api.interval(hours=2).to_unit('m'), api.interval(minutes=2 * 60)),
        (api.interval(hours=2).to_unit('s'), api.interval(seconds=2 * 3600)),
        (
            api.interval(hours=2).to_unit('ms'),
            api.interval(milliseconds=2 * 3600000),
        ),
        (
            api.interval(hours=2).to_unit('us'),
            api.interval(microseconds=2 * 3600000000),
        ),
        (
            api.interval(hours=2).to_unit('ns'),
            api.interval(nanoseconds=2 * 3600000000000),
        ),
    ],
)
def test_downconvert_hours(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.type().unit == expected.type().unit


@pytest.mark.parametrize(
    ('case', 'expected'),
    [
        (api.interval(weeks=2).to_unit('D'), api.interval(days=2 * 7)),
        (api.interval(weeks=2).to_unit('h'), api.interval(hours=2 * 7 * 24)),
        (api.interval(days=2).to_unit('D'), api.interval(days=2)),
        (api.interval(days=2).to_unit('h'), api.interval(hours=2 * 24)),
        (api.interval(days=2).to_unit('m'), api.interval(minutes=2 * 1440)),
        (api.interval(days=2).to_unit('s'), api.interval(seconds=2 * 86400)),
        (
            api.interval(days=2).to_unit('ms'),
            api.interval(milliseconds=2 * 86400000),
        ),
        (
            api.interval(days=2).to_unit('us'),
            api.interval(microseconds=2 * 86400000000),
        ),
        (
            api.interval(days=2).to_unit('ns'),
            api.interval(nanoseconds=2 * 86400000000000),
        ),
    ],
)
def test_downconvert_day(case, expected):
    assert isinstance(case, ir.IntervalScalar)
    assert isinstance(expected, ir.IntervalScalar)
    assert case.type().unit == expected.type().unit


@pytest.mark.parametrize(
    ('a', 'b', 'unit'),
    [
        (api.interval(days=1), api.interval(days=3), 'D'),
        (api.interval(seconds=1), api.interval(hours=10), 's'),
        (api.interval(hours=3), api.interval(days=2), 'h'),
    ],
)
def test_combine_with_different_kinds(a, b, unit):
    assert (a + b).type().unit == unit


@pytest.mark.parametrize(
    ('case', 'expected'),
    [
        (api.interval(quarters=2), api.interval(quarters=2)),
        (api.interval(weeks=2), api.interval(weeks=2)),
        (api.interval(days=3), api.interval(days=3)),
        (api.interval(hours=4), api.interval(hours=4)),
        (api.interval(minutes=5), api.interval(minutes=5)),
        (api.interval(seconds=6), api.interval(seconds=6)),
        (api.interval(milliseconds=7), api.interval(milliseconds=7)),
        (api.interval(microseconds=8), api.interval(microseconds=8)),
        (api.interval(nanoseconds=9), api.interval(nanoseconds=9)),
    ],
)
def test_timedelta_generic_api(case, expected):
    assert case.equals(expected)


def test_interval_time_expr(table):
    c = table.k
    x = api.interval(hours=1)

    expr = x + c
    assert isinstance(expr, ir.TimeColumn)
    assert isinstance(expr.op(), ops.TimeAdd)

    # test radd
    expr = c + x
    assert isinstance(expr, ir.TimeColumn)
    assert isinstance(expr.op(), ops.TimeAdd)

    expr = x - c
    assert isinstance(expr, ir.TimeColumn)
    assert isinstance(expr.op(), ops.TimeSub)

    # test radd
    expr = c - x
    assert isinstance(expr, ir.TimeColumn)
    assert isinstance(expr.op(), ops.TimeSub)


def test_interval_date_expr(table):
    c = table.j
    x = api.interval(days=1)

    expr = x + c
    assert isinstance(expr, ir.DateColumn)
    assert isinstance(expr.op(), ops.DateAdd)

    # test radd
    expr = c + x
    assert isinstance(expr, ir.DateColumn)
    assert isinstance(expr.op(), ops.DateAdd)

    expr = x - c
    assert isinstance(expr, ir.DateColumn)
    assert isinstance(expr.op(), ops.DateSub)

    # test radd
    expr = c - x
    assert isinstance(expr, ir.DateColumn)
    assert isinstance(expr.op(), ops.DateSub)


def test_interval_timestamp_expr(table):
    c = table.i
    x = api.interval(seconds=1)

    expr = x + c
    assert isinstance(expr, ir.TimestampColumn)
    assert isinstance(expr.op(), ops.TimestampAdd)

    # test radd
    expr = c + x
    assert isinstance(expr, ir.TimestampColumn)
    assert isinstance(expr.op(), ops.TimestampAdd)

    expr = x - c
    assert isinstance(expr, ir.TimestampColumn)
    assert isinstance(expr.op(), ops.TimestampSub)

    # test radd
    expr = c - x
    assert isinstance(expr, ir.TimestampColumn)
    assert isinstance(expr.op(), ops.TimestampSub)


@pytest.mark.xfail(raises=AssertionError, reason='NYI')
def test_compound_offset():
    # These are not yet allowed (e.g. 1 month + 1 hour)
    assert False


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_offset_months():
    assert False


@pytest.mark.parametrize(
    'literal',
    [
        api.interval(3600),
        api.interval(datetime.timedelta(days=3)),
        api.interval(years=1),
        api.interval(quarters=3),
        api.interval(months=2),
        api.interval(weeks=3),
        api.interval(days=-1),
        api.interval(hours=-3),
        api.interval(minutes=5),
        api.interval(seconds=-8),
    ],
)
def test_interval(literal):
    assert isinstance(literal, ir.IntervalScalar)


@pytest.mark.parametrize(
    ('expr', 'expected'),
    [
        (api.interval(weeks=3), "Literal[interval<int8>(unit='W')]\n  3"),
        (api.interval(months=3), "Literal[interval<int8>(unit='M')]\n  3"),
        (
            api.interval(seconds=-10),
            "Literal[interval<int8>(unit='s')]\n  -10",
        ),
    ],
)
def test_interval_repr(expr, expected):
    assert repr(expr) == expected


def test_timestamp_arithmetics():
    ts1 = api.timestamp(datetime.datetime.now())
    ts2 = api.timestamp(datetime.datetime.today())

    i1 = api.interval(minutes=30)

    # TODO: raise for unsupported operations too
    for expr in [ts2 - ts1, ts1 - ts2]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.TimestampDiff)
        assert expr.type() == dt.Interval('s', dt.int32)

    for expr in [ts1 - i1, ts2 - i1]:
        assert isinstance(expr, ir.TimestampScalar)
        assert isinstance(expr.op(), ops.TimestampSub)

    for expr in [ts1 + i1, ts2 + i1]:
        assert isinstance(expr, ir.TimestampScalar)
        assert isinstance(expr.op(), ops.TimestampAdd)


def test_date_arithmetics():
    d1 = api.date('2015-01-02')
    d2 = api.date('2017-01-01')
    i1 = api.interval(weeks=3)

    for expr in [d1 - d2, d2 - d1]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.DateDiff)
        assert expr.type() == dt.Interval('D', dt.int32)

    for expr in [d1 - i1, d2 - i1]:
        assert isinstance(expr, ir.DateScalar)
        assert isinstance(expr.op(), ops.DateSub)

    for expr in [d1 + i1, d2 + i1]:
        assert isinstance(expr, ir.DateScalar)
        assert isinstance(expr.op(), ops.DateAdd)


def test_time_arithmetics():
    t1 = api.time('18:00')
    t2 = api.time('19:12')
    i1 = api.interval(minutes=3)

    for expr in [t1 - t2, t2 - t1]:
        assert isinstance(expr, ir.IntervalScalar)
        assert isinstance(expr.op(), ops.TimeDiff)
        assert expr.type() == dt.Interval('s', dt.int32)

    for expr in [t1 - i1, t2 - i1]:
        assert isinstance(expr, ir.TimeScalar)
        assert isinstance(expr.op(), ops.TimeSub)

    for expr in [t1 + i1, t2 + i1]:
        assert isinstance(expr, ir.TimeScalar)
        assert isinstance(expr.op(), ops.TimeAdd)


def test_invalid_date_arithmetics():
    d1 = api.date('2015-01-02')
    i1 = api.interval(seconds=300)
    i2 = api.interval(minutes=15)
    i3 = api.interval(hours=1)

    for i in [i1, i2, i3]:
        with pytest.raises(TypeError):
            d1 - i
        with pytest.raises(TypeError):
            d1 + i


@pytest.mark.parametrize(
    ('prop', 'expected_unit'),
    [
        ('nanoseconds', 'ns'),
        ('microseconds', 'us'),
        ('milliseconds', 'ms'),
        ('seconds', 's'),
        ('minutes', 'm'),
        ('hours', 'h'),
        ('days', 'D'),
        ('weeks', 'W'),
    ],
)
def test_interval_properties(prop, expected_unit):
    i = api.interval(seconds=3600)
    assert getattr(i, prop).type().unit == expected_unit


@pytest.mark.parametrize(
    'interval',
    [api.interval(years=1), api.interval(quarters=4), api.interval(months=12)],
)
@pytest.mark.parametrize(
    ('prop', 'expected_unit'),
    [('months', 'M'), ('quarters', 'Q'), ('years', 'Y')],
)
def test_interval_monthly_properties(interval, prop, expected_unit):
    assert getattr(interval, prop).type().unit == expected_unit


@pytest.mark.parametrize(
    ('interval', 'prop'),
    [
        (api.interval(hours=48), 'months'),
        (api.interval(years=2), 'seconds'),
        (api.interval(quarters=1), 'weeks'),
    ],
)
def test_unsupported_properties(interval, prop):
    with pytest.raises(ValueError):
        getattr(interval, prop)


@pytest.mark.parametrize('column', ['a', 'b', 'c', 'd'])  # integer columns
@pytest.mark.parametrize(
    'unit', ['Y', 'Q', 'M', 'D', 'W', 'h', 'm', 's', 'ms', 'us', 'ns']
)
def test_integer_to_interval(column, unit, table):
    c = table[column]
    i = c.to_interval(unit)
    assert isinstance(i, ir.IntervalColumn)
    assert i.type().value_type == c.type()
    assert i.type().unit == unit


@pytest.mark.parametrize(
    'unit', ['Y', 'Q', 'M', 'D', 'W', 'h', 'm', 's', 'ms', 'us', 'ns']
)
@pytest.mark.parametrize(
    'operands',
    [
        lambda t, u: (api.interval(3, unit=u), api.interval(2, unit=u)),
        lambda t, u: (api.interval(3, unit=u), api.interval(3, unit=u)),
        lambda t, u: (t.c.to_interval(unit=u), api.interval(2, unit=u)),
        lambda t, u: (t.c.to_interval(unit=u), t.d.to_interval(unit=u)),
    ],
)
@pytest.mark.parametrize(
    'operator',
    [
        operator.eq,
        operator.ne,
        operator.ge,
        operator.gt,
        operator.le,
        operator.lt,
    ],
    ids=lambda op: op.__name__,
)
def test_interval_comparisons(unit, operands, operator, table):
    a, b = operands(table, unit)
    expr = operator(a, b)

    assert isinstance(a, ir.IntervalValue)
    assert isinstance(b, ir.IntervalValue)
    assert isinstance(expr, ir.BooleanValue)


@pytest.mark.parametrize(
    'operands',
    [
        lambda t: (api.date('2016-01-01'), api.date('2016-02-02')),
        lambda t: (t.j, api.date('2016-01-01')),
        lambda t: (api.date('2016-01-01'), t.j),
        lambda t: (t.j, t.i.date()),
        lambda t: (t.i.date(), t.j),
    ],
    ids=[
        'literal-literal',
        'column-literal',
        'literal-column',
        'column-casted',
        'casted-column',
    ],
)
@pytest.mark.parametrize(
    'interval',
    [
        lambda t: api.interval(years=4),
        lambda t: api.interval(quarters=4),
        lambda t: api.interval(months=3),
        lambda t: api.interval(weeks=2),
        lambda t: api.interval(days=1),
        lambda t: t.c.to_interval(unit='Y'),
        lambda t: t.c.to_interval(unit='M'),
        lambda t: t.c.to_interval(unit='W'),
        lambda t: t.c.to_interval(unit='D'),
    ],
    ids=[
        'years',
        'quarters',
        'months',
        'weeks',
        'days',
        'to-years',
        'to-months',
        'to-weeks',
        'to-days',
    ],
)
@pytest.mark.parametrize(
    'arithmetic',
    [lambda a, i: a - i, lambda a, i: a + i, lambda a, i: i + a],
    ids=['subtract', 'radd', 'add'],
)
@pytest.mark.parametrize(
    'operator',
    [
        operator.eq,
        operator.ne,
        operator.ge,
        operator.gt,
        operator.le,
        operator.lt,
    ],
    ids=lambda op: op.__name__,
)
def test_complex_date_comparisons(
    operands, interval, arithmetic, operator, table
):
    (a, b), i = operands(table), interval(table)

    a_ = arithmetic(a, i)
    expr = operator(a_, b)

    assert isinstance(a, ir.DateValue)
    assert isinstance(b, ir.DateValue)
    assert isinstance(a_, ir.DateValue)
    assert isinstance(i, ir.IntervalValue)
    assert isinstance(expr, ir.BooleanValue)


def test_interval_column_name(table):
    c = table.i
    expr = (c - c).name('foo')
    assert expr._name == 'foo'


@pytest.mark.parametrize(
    'operand',
    [lambda t: api.timestamp(datetime.datetime.now()), lambda t: t.i],
    ids=['column', 'literal'],
)
@pytest.mark.parametrize(
    'unit',
    [
        'Y',
        'y',
        'year',
        'YEAR',
        'YYYY',
        'SYYYY',
        'YYY',
        'YY',
        'Q',
        'q',
        'quarter',
        'QUARTER',
        'M',
        'month',
        'MONTH',
        'w',
        'W',
        'week',
        'WEEK',
        'd',
        'J',
        'day',
        'DAY',
        'h',
        'H',
        'HH24',
        'hour',
        'HOUR',
        'm',
        'MI',
        'minute',
        'MINUTE',
        's',
        'second',
        'SECOND',
        'ms',
        'millisecond',
        'MILLISECOND',
        'us',
        'microsecond',
        'MICROSECOND',
        'ns',
        'nanosecond',
        'NANOSECOND',
    ],
)
def test_timestamp_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.TimestampValue)
    assert isinstance(expr.op(), ops.TimestampTruncate)


@pytest.mark.parametrize(
    'operand', [lambda t: api.date('2018-01-01'), lambda t: t.j]
)
@pytest.mark.parametrize('unit', ['Y', 'Q', 'M', 'D', 'W'])
def test_date_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.DateValue)
    assert isinstance(expr.op(), ops.DateTruncate)


@pytest.mark.parametrize(
    'operand', [lambda t: api.time('18:00'), lambda t: t.k]
)
@pytest.mark.parametrize('unit', ['h', 'm', 's', 'ms', 'us', 'ns'])
def test_time_truncate(table, operand, unit):
    expr = operand(table).truncate(unit)
    assert isinstance(expr, ir.TimeValue)
    assert isinstance(expr.op(), ops.TimeTruncate)



import enum

import pytest
from toolz import identity

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.common.exceptions import IbisTypeError

table = ibis.table(
    [('int_col', 'int64'), ('string_col', 'string'), ('double_col', 'double')]
)


@pytest.mark.parametrize(
    ('value', 'expected'),
    [
        (dt.int32, dt.int32),
        ('int64', dt.int64),
        ('array<string>', dt.Array(dt.string)),
    ],
)
def test_valid_datatype(value, expected):
    assert rlz.datatype(value) == expected


@pytest.mark.parametrize(
    ('value', 'expected'),
    [
        ('exception', IbisTypeError),
        ('array<cat>', IbisTypeError),
        (int, IbisTypeError),
        ([float], IbisTypeError),
    ],
)
def test_invalid_datatype(value, expected):
    with pytest.raises(expected):
        assert rlz.datatype(value)


@pytest.mark.parametrize(
    ('klass', 'value', 'expected'),
    [(int, 32, 32), (str, 'foo', 'foo'), (dt.Integer, dt.int8, dt.int8)],
)
def test_valid_instance_of(klass, value, expected):
    assert rlz.instance_of(klass, value) == expected


@pytest.mark.parametrize(
    ('klass', 'value', 'expected'),
    [
        (ir.TableExpr, object, IbisTypeError),
        (ir.IntegerValue, 4, IbisTypeError),
    ],
)
def test_invalid_instance_of(klass, value, expected):
    with pytest.raises(expected):
        assert rlz.instance_of(klass, value)


@pytest.mark.parametrize(
    ('dtype', 'value', 'expected'),
    [
        pytest.param(dt.int8, 26, ibis.literal(26)),
        pytest.param(dt.int16, 26, ibis.literal(26)),
        pytest.param(dt.int32, 26, ibis.literal(26)),
        pytest.param(dt.int64, 26, ibis.literal(26)),
        pytest.param(dt.uint8, 26, ibis.literal(26)),
        pytest.param(dt.uint16, 26, ibis.literal(26)),
        pytest.param(dt.uint32, 26, ibis.literal(26)),
        pytest.param(dt.uint64, 26, ibis.literal(26)),
        pytest.param(dt.float32, 26, ibis.literal(26)),
        pytest.param(dt.float64, 26.4, ibis.literal(26.4)),
        pytest.param(dt.double, 26.3, ibis.literal(26.3)),
        pytest.param(dt.string, 'bar', ibis.literal('bar')),
        pytest.param(dt.Array(dt.float), [3.4, 5.6], ibis.literal([3.4, 5.6])),
        pytest.param(
            dt.Map(dt.string, dt.Array(dt.boolean)),
            {'a': [True, False], 'b': [True]},
            ibis.literal({'a': [True, False], 'b': [True]}),
            id='map_literal',
        ),
    ],
)
def test_valid_value(dtype, value, expected):
    result = rlz.value(dtype, value)
    assert result.equals(expected)


@pytest.mark.parametrize(
    ('dtype', 'value', 'expected'),
    [
        (dt.uint8, -3, IbisTypeError),
        (dt.int32, dict(), IbisTypeError),
        (dt.string, 1, IbisTypeError),
        (dt.Array(dt.float), ['s'], IbisTypeError),
        (
            dt.Map(dt.string, dt.Array(dt.boolean)),
            {'a': [True, False], 'b': ['B']},
            IbisTypeError,
        ),
    ],
)
def test_invalid_value(dtype, value, expected):
    with pytest.raises(expected):
        rlz.value(dtype, value)


@pytest.mark.parametrize(
    ('values', 'value', 'expected'),
    [
        (['a', 'b'], 'a', 'a'),
        (('a', 'b'), 'b', 'b'),
        ({'a', 'b', 'c'}, 'c', 'c'),
        ([1, 2, 'f'], 'f', 'f'),
        ({'a': 1, 'b': 2}, 'a', 1),
        ({'a': 1, 'b': 2}, 'b', 2),
    ],
)
def test_valid_isin(values, value, expected):
    assert rlz.isin(values, value) == expected


@pytest.mark.parametrize(
    ('values', 'value', 'expected'),
    [
        (['a', 'b'], 'c', ValueError),
        ({'a', 'b', 'c'}, 'd', ValueError),
        ({'a': 1, 'b': 2}, 'c', ValueError),
    ],
)
def test_invalid_isin(values, value, expected):
    with pytest.raises(expected):
        rlz.isin(values, value)


class Foo(enum.Enum):
    a = 1
    b = 2


class Bar:
    a = 'A'
    b = 'B'


class Baz:
    def __init__(self, a):
        self.a = a


@pytest.mark.parametrize(
    ('obj', 'value', 'expected'),
    [
        (Foo, Foo.a, Foo.a),
        (Foo, 'b', Foo.b),
        (Bar, 'a', 'A'),
        (Bar, 'b', 'B'),
        (Baz(2), 'a', 2),
    ],
)
def test_valid_member_of(obj, value, expected):
    assert rlz.member_of(obj, value) == expected


@pytest.mark.parametrize(
    ('obj', 'value', 'expected'),
    [
        (Foo, 'c', IbisTypeError),
        (Bar, 'c', IbisTypeError),
        (Baz(3), 'b', IbisTypeError),
    ],
)
def test_invalid_member_of(obj, value, expected):
    with pytest.raises(expected):
        rlz.member_of(obj, value)


@pytest.mark.parametrize(
    ('validator', 'values', 'expected'),
    [
        (rlz.list_of(identity), (3, 2), ibis.sequence([3, 2])),
        (rlz.list_of(rlz.integer), (3, 2), ibis.sequence([3, 2])),
        (rlz.list_of(rlz.integer), (3, None), ibis.sequence([3, ibis.NA])),
        (rlz.list_of(rlz.string), ('a',), ibis.sequence(['a'])),
        (rlz.list_of(rlz.string), ['a', 'b'], ibis.sequence(['a', 'b'])),
        pytest.param(
            rlz.list_of(rlz.list_of(rlz.string)),
            [[], ['a']],
            ibis.sequence([[], ['a']]),
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Not yet implemented'
            ),
        ),
        (
            rlz.list_of(rlz.boolean, min_length=2),
            [True, False],
            ibis.sequence([True, False]),
        ),
    ],
)
def test_valid_list_of(validator, values, expected):
    result = validator(values)
    assert result.equals(expected)


@pytest.mark.parametrize(
    ('validator', 'values'),
    [
        (rlz.list_of(rlz.double, min_length=2), [1]),
        (rlz.list_of(rlz.integer), 1.1),
        (rlz.list_of(rlz.string), 'asd'),
        (rlz.list_of(identity), 3),
    ],
)
def test_invalid_list_of(validator, values):
    with pytest.raises(IbisTypeError):
        validator(values)


@pytest.mark.parametrize(
    ('units', 'value', 'expected'),
    [
        ({'H', 'D'}, ibis.interval(days=3), ibis.interval(days=3)),
        (['Y'], ibis.interval(years=3), ibis.interval(years=3)),
    ],
)
def test_valid_interval(units, value, expected):
    result = rlz.interval(value, units=units)
    assert result.equals(expected)


@pytest.mark.parametrize(
    ('units', 'value', 'expected'),
    [
        ({'Y'}, ibis.interval(hours=1), IbisTypeError),
        ({'Y', 'M', 'D'}, ibis.interval(hours=1), IbisTypeError),
        ({'Q', 'W', 'D'}, ibis.interval(seconds=1), IbisTypeError),
    ],
)
def test_invalid_interval(units, value, expected):
    with pytest.raises(expected):
        rlz.interval(value, units=units)


@pytest.mark.parametrize(
    ('validator', 'value', 'expected'),
    [
        (rlz.column(rlz.any), table.int_col, table.int_col),
        (rlz.column(rlz.string), table.string_col, table.string_col),
        (rlz.scalar(rlz.integer), ibis.literal(3), ibis.literal(3)),
        (rlz.scalar(rlz.any), 'caracal', ibis.literal('caracal')),
    ],
)
def test_valid_column_or_scalar(validator, value, expected):
    result = validator(value)
    assert result.equals(expected)


@pytest.mark.parametrize(
    ('validator', 'value', 'expected'),
    [
        (rlz.column(rlz.integer), table.double_col, IbisTypeError),
        (rlz.column(rlz.any), ibis.literal(3), IbisTypeError),
        (rlz.column(rlz.integer), ibis.literal(3), IbisTypeError),
    ],
)
def test_invalid_column_or_scalar(validator, value, expected):
    with pytest.raises(expected):
        validator(value)


@pytest.mark.parametrize(
    'table',
    [
        ibis.table([('group', dt.int64), ('value', dt.double)]),
        ibis.table(
            [('group', dt.int64), ('value', dt.double), ('value2', dt.double)]
        ),
    ],
)
def test_table_with_schema(table):
    validator = rlz.table([('group', dt.int64), ('value', dt.double)])
    assert validator(table) == table


@pytest.mark.parametrize(
    'table', [ibis.table([('group', dt.int64), ('value', dt.timestamp)])]
)
def test_table_with_schema_invalid(table):
    validator = rlz.table([('group', dt.double), ('value', dt.timestamp)])
    with pytest.raises(IbisTypeError):
        validator(table)


def test_shape_like_with_no_arguments():
    with pytest.raises(ValueError) as e:
        rlz.shape_like([])
    assert str(e.value) == 'Must pass at least one expression'


@pytest.mark.parametrize(
    ('rule', 'input'),
    [
        (rlz.array_of(rlz.integer), [1, 2, 3]),
        (rlz.array_of(rlz.integer), []),
        (rlz.array_of(rlz.double), [1, 2]),
        (rlz.array_of(rlz.string), ['a', 'b']),
        (rlz.array_of(rlz.array_of(rlz.string)), [['a'], [], [], ['a', 'b']]),
    ],
)
def test_array_of(rule, input):
    assert isinstance(rule(input).type(), dt.Array)


@pytest.mark.parametrize(
    ('rule', 'input'),
    [
        (rlz.array_of(rlz.array_of(rlz.string)), [1, 2]),
        (rlz.array_of(rlz.string), [1, 2.0]),
        (rlz.array_of(rlz.array_of(rlz.integer)), [2, 2.0]),
    ],
)
def test_array_of_invalid_input(rule, input):
    with pytest.raises(IbisTypeError):
        rule(input)



import pytest
from pkg_resources import parse_version

import ibis
import ibis.expr.datatypes as dt
from ibis.tests.backends import BigQuery


@pytest.mark.xfail_unsupported
def test_version(backend, con):
    expected_type = (
        type(parse_version('1.0')),
        type(parse_version('1.0-legacy')),
    )
    assert isinstance(con.version, expected_type)


@pytest.mark.parametrize(
    ('expr_fn', 'expected'),
    [
        (lambda t: t.string_col, [('string_col', dt.String)]),
        (
            lambda t: t[t.string_col, t.bigint_col],
            [('string_col', dt.String), ('bigint_col', dt.Int64)],
        ),
    ],
)
def test_query_schema(backend, con, alltypes, expr_fn, expected):
    if not hasattr(con, '_build_ast'):
        pytest.skip(
            '{} backend has no _build_ast method'.format(
                type(backend).__name__
            )
        )

    expr = expr_fn(alltypes)

    # we might need a public API for it
    ast = con._build_ast(expr, backend.make_context())
    query = con.query_class(con, ast)
    schema = query.schema()

    # clickhouse columns has been defined as non-nullable
    # whereas other backends don't support non-nullable columns yet
    expected = ibis.schema(
        [
            (name, dtype(nullable=schema[name].nullable))
            for name, dtype in expected
        ]
    )
    assert query.schema().equals(expected)


@pytest.mark.parametrize(
    'sql',
    [
        'select * from functional_alltypes limit 10',
        'select * from functional_alltypes \nlimit 10\n',
    ],
)
@pytest.mark.xfail_backends((BigQuery,))
@pytest.mark.xfail_unsupported
def test_sql(backend, con, sql):
    if not hasattr(con, 'sql') or not hasattr(con, '_get_schema_using_query'):
        pytest.skip('Backend {} does not support sql method'.format(backend))

    # execute the expression using SQL query
    con.sql(sql).execute()
